from __future__ import annotations

from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Literal,
    Protocol,
    TypeAlias,
    TypeVar,
    cast,
)

if TYPE_CHECKING:
    from .event_stream import EventStream


ThinkingLevel: TypeAlias = Literal["off", "minimal", "low", "medium", "high", "xhigh"]
Transport: TypeAlias = Literal["sse", "websocket", "auto"]
StopReason: TypeAlias = Literal["stop", "length", "toolUse", "aborted", "error"]


def now_ms() -> int:
    import time

    return int(time.time() * 1000)


@dataclass
class UsageCost:
    input: float = 0.0
    output: float = 0.0
    cache_read: float = 0.0
    cache_write: float = 0.0
    total: float = 0.0


@dataclass
class Usage:
    input: int = 0
    output: int = 0
    cache_read: int = 0
    cache_write: int = 0
    total_tokens: int = 0
    cost: UsageCost = field(default_factory=UsageCost)


@dataclass
class TextContent:
    text: str
    type: Literal["text"] = "text"


@dataclass
class ImageContent:
    data: str
    mime_type: str
    type: Literal["image"] = "image"


@dataclass
class ThinkingContent:
    thinking: str
    thinking_signature: str | None = None
    type: Literal["thinking"] = "thinking"


@dataclass
class ToolCallContent:
    id: str
    name: str
    arguments: dict[str, Any] = field(default_factory=dict)
    type: Literal["toolCall"] = "toolCall"


UserContentPart: TypeAlias = TextContent | ImageContent
AssistantContentPart: TypeAlias = TextContent | ThinkingContent | ToolCallContent
ToolResultContentPart: TypeAlias = TextContent | ImageContent


@dataclass
class UserMessage:
    content: str | list[UserContentPart]
    timestamp: int = field(default_factory=now_ms)
    role: Literal["user"] = "user"


@dataclass
class AssistantMessage:
    content: list[AssistantContentPart]
    api: str
    provider: str
    model: str
    usage: Usage = field(default_factory=Usage)
    stop_reason: StopReason = "stop"
    timestamp: int = field(default_factory=now_ms)
    error_message: str | None = None
    role: Literal["assistant"] = "assistant"


@dataclass
class ToolResultMessage:
    tool_call_id: str
    tool_name: str
    content: list[ToolResultContentPart]
    details: Any = None
    is_error: bool = False
    timestamp: int = field(default_factory=now_ms)
    role: Literal["toolResult"] = "toolResult"


Message: TypeAlias = UserMessage | AssistantMessage | ToolResultMessage
AgentMessage: TypeAlias = Message | dict[str, Any] | Any


@dataclass
class Model:
    id: str
    name: str
    api: str
    provider: str
    base_url: str | None = None
    reasoning: bool = False
    max_tokens: int | None = None


ThinkingBudgets: TypeAlias = dict[Literal["minimal", "low", "medium", "high"], int]


class AbortSignal(Protocol):
    @property
    def aborted(self) -> bool:
        ...


@dataclass
class SimpleStreamOptions:
    api_key: str | None = None
    temperature: float | None = None
    max_tokens: int | None = None
    reasoning: ThinkingLevel | None = None
    signal: AbortSignal | None = None
    session_id: str | None = None
    thinking_budgets: ThinkingBudgets | None = None
    transport: Transport = "sse"
    max_retry_delay_ms: int | None = None


@dataclass
class AgentToolResult:
    content: list[ToolResultContentPart]
    details: Any


AgentToolUpdateCallback: TypeAlias = Callable[[AgentToolResult], None]
AgentToolExecute: TypeAlias = Callable[
    [str, dict[str, Any], AbortSignal | None, AgentToolUpdateCallback | None],
    Awaitable[AgentToolResult],
]


@dataclass
class AgentTool:
    name: str
    label: str
    description: str
    execute: AgentToolExecute
    parameters: Any = None
    validate: Callable[[dict[str, Any]], dict[str, Any]] | None = None


@dataclass
class Context:
    system_prompt: str
    messages: list[Message]
    tools: list[AgentTool] | None = None


@dataclass
class AgentContext:
    system_prompt: str
    messages: list[AgentMessage]
    tools: list[AgentTool] | None = None


ConvertToLlm: TypeAlias = Callable[[list[AgentMessage]], list[Message] | Awaitable[list[Message]]]
TransformContext: TypeAlias = Callable[[list[AgentMessage], AbortSignal | None], Awaitable[list[AgentMessage]]]
QueuedMessagesCallback: TypeAlias = Callable[[], list[AgentMessage] | Awaitable[list[AgentMessage]]]
ApiKeyCallback: TypeAlias = Callable[[str], str | None | Awaitable[str | None]]


@dataclass
class AgentLoopConfig(SimpleStreamOptions):
    model: Model | None = None
    convert_to_llm: ConvertToLlm | None = None
    transform_context: TransformContext | None = None
    get_api_key: ApiKeyCallback | None = None
    get_steering_messages: QueuedMessagesCallback | None = None
    get_follow_up_messages: QueuedMessagesCallback | None = None


# Assistant stream events
@dataclass
class AssistantStartEvent:
    partial: AssistantMessage
    type: Literal["start"] = "start"


@dataclass
class AssistantTextStartEvent:
    content_index: int
    partial: AssistantMessage
    type: Literal["text_start"] = "text_start"


@dataclass
class AssistantTextDeltaEvent:
    content_index: int
    delta: str
    partial: AssistantMessage
    type: Literal["text_delta"] = "text_delta"


@dataclass
class AssistantTextEndEvent:
    content_index: int
    content: str
    partial: AssistantMessage
    type: Literal["text_end"] = "text_end"


@dataclass
class AssistantThinkingStartEvent:
    content_index: int
    partial: AssistantMessage
    type: Literal["thinking_start"] = "thinking_start"


@dataclass
class AssistantThinkingDeltaEvent:
    content_index: int
    delta: str
    partial: AssistantMessage
    type: Literal["thinking_delta"] = "thinking_delta"


@dataclass
class AssistantThinkingEndEvent:
    content_index: int
    content: str
    partial: AssistantMessage
    type: Literal["thinking_end"] = "thinking_end"


@dataclass
class AssistantToolCallStartEvent:
    content_index: int
    partial: AssistantMessage
    type: Literal["toolcall_start"] = "toolcall_start"


@dataclass
class AssistantToolCallDeltaEvent:
    content_index: int
    delta: str
    partial: AssistantMessage
    type: Literal["toolcall_delta"] = "toolcall_delta"


@dataclass
class AssistantToolCallEndEvent:
    content_index: int
    tool_call: ToolCallContent
    partial: AssistantMessage
    type: Literal["toolcall_end"] = "toolcall_end"


@dataclass
class AssistantDoneEvent:
    reason: Literal["stop", "length", "toolUse"]
    message: AssistantMessage
    type: Literal["done"] = "done"


@dataclass
class AssistantErrorEvent:
    reason: Literal["aborted", "error"]
    error: AssistantMessage
    type: Literal["error"] = "error"


AssistantMessageEvent: TypeAlias = (
    AssistantStartEvent
    | AssistantTextStartEvent
    | AssistantTextDeltaEvent
    | AssistantTextEndEvent
    | AssistantThinkingStartEvent
    | AssistantThinkingDeltaEvent
    | AssistantThinkingEndEvent
    | AssistantToolCallStartEvent
    | AssistantToolCallDeltaEvent
    | AssistantToolCallEndEvent
    | AssistantDoneEvent
    | AssistantErrorEvent
)


# Agent events
@dataclass
class AgentStartEvent:
    type: Literal["agent_start"] = "agent_start"


@dataclass
class AgentEndEvent:
    messages: list[AgentMessage]
    type: Literal["agent_end"] = "agent_end"


@dataclass
class TurnStartEvent:
    type: Literal["turn_start"] = "turn_start"


@dataclass
class TurnEndEvent:
    message: AgentMessage
    tool_results: list[ToolResultMessage]
    type: Literal["turn_end"] = "turn_end"


@dataclass
class MessageStartEvent:
    message: AgentMessage
    type: Literal["message_start"] = "message_start"


@dataclass
class MessageUpdateEvent:
    message: AgentMessage
    assistant_message_event: AssistantMessageEvent
    type: Literal["message_update"] = "message_update"


@dataclass
class MessageEndEvent:
    message: AgentMessage
    type: Literal["message_end"] = "message_end"


@dataclass
class ToolExecutionStartEvent:
    tool_call_id: str
    tool_name: str
    args: Any
    type: Literal["tool_execution_start"] = "tool_execution_start"


@dataclass
class ToolExecutionUpdateEvent:
    tool_call_id: str
    tool_name: str
    args: Any
    partial_result: AgentToolResult
    type: Literal["tool_execution_update"] = "tool_execution_update"


@dataclass
class ToolExecutionEndEvent:
    tool_call_id: str
    tool_name: str
    result: AgentToolResult
    is_error: bool
    type: Literal["tool_execution_end"] = "tool_execution_end"


AgentEvent: TypeAlias = (
    AgentStartEvent
    | AgentEndEvent
    | TurnStartEvent
    | TurnEndEvent
    | MessageStartEvent
    | MessageUpdateEvent
    | MessageEndEvent
    | ToolExecutionStartEvent
    | ToolExecutionUpdateEvent
    | ToolExecutionEndEvent
)


class _AwaitableStream(Protocol):
    def __await__(self):
        ...


StreamFn: TypeAlias = Callable[
    [Model, Context, SimpleStreamOptions],
    "EventStream[AssistantMessageEvent, AssistantMessage]"
    | Awaitable["EventStream[AssistantMessageEvent, AssistantMessage]"],
]


@dataclass
class AgentState:
    system_prompt: str = ""
    model: Model = field(
        default_factory=lambda: Model(
            id="mock-model",
            name="Mock Model",
            api="mock-api",
            provider="mock-provider",
        )
    )
    thinking_level: ThinkingLevel = "off"
    tools: list[AgentTool] = field(default_factory=list)
    messages: list[AgentMessage] = field(default_factory=list)
    is_streaming: bool = False
    stream_message: AgentMessage | None = None
    pending_tool_calls: set[str] = field(default_factory=set)
    error: str | None = None


def message_role(message: AgentMessage) -> str:
    if isinstance(message, dict):
        return str(message.get("role", ""))
    return str(getattr(message, "role", ""))


def is_assistant_message(message: AgentMessage) -> bool:
    return message_role(message) == "assistant"


def as_assistant_message(message: AgentMessage) -> AssistantMessage:
    if not isinstance(message, AssistantMessage):
        raise TypeError("Expected AssistantMessage")
    return message


T = TypeVar("T")


def coerce_awaitable(value: T | Awaitable[T]) -> Awaitable[T]:
    import inspect

    if inspect.isawaitable(value):
        return cast(Awaitable[T], value)

    async def _wrap() -> T:
        return value

    return _wrap()


def validate_tool_arguments(tool: AgentTool, tool_call: ToolCallContent) -> dict[str, Any]:
    args = dict(tool_call.arguments)
    if tool.validate:
        return tool.validate(args)
    return args



"""fluxos-agent - Stateful agent loop and runtime for Python."""

from .agent import Agent, AgentOptions
from .agent_loop import agent_loop, agent_loop_continue
from .event_stream import EventStream
from .proxy import ProxyMessageEventStream, ProxyStreamOptions, stream_proxy
from .types import (
    AgentContext,
    AgentLoopConfig,
    AgentMessage,
    AgentState,
    AgentTool,
    AgentToolResult,
    AssistantMessage,
    AssistantMessageEvent,
    Context,
    ImageContent,
    Message,
    Model,
    TextContent,
    ThinkingBudgets,
    ThinkingContent,
    ThinkingLevel,
    ToolCallContent,
    ToolResultMessage,
)
from .version import __version__

__all__ = [
    "Agent",
    "AgentOptions",
    "EventStream",
    "agent_loop",
    "agent_loop_continue",
    "stream_proxy",
    "ProxyMessageEventStream",
    "ProxyStreamOptions",
    "Model",
    "Context",
    "Message",
    "AgentMessage",
    "AssistantMessage",
    "ToolResultMessage",
    "AgentContext",
    "AgentLoopConfig",
    "AgentState",
    "AgentTool",
    "AgentToolResult",
    "TextContent",
    "ImageContent",
    "ThinkingContent",
    "ToolCallContent",
    "AssistantMessageEvent",
    "ThinkingLevel",
    "ThinkingBudgets",
    "__version__",
]



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import Any, Callable

from .agent_loop import agent_loop, agent_loop_continue, build_error_message
from .types import (
    AgentContext,
    AgentEndEvent,
    AgentEvent,
    AgentLoopConfig,
    AgentMessage,
    AgentState,
    AgentTool,
    ImageContent,
    Message,
    Model,
    TextContent,
    ThinkingBudgets,
    ThinkingLevel,
    Transport,
    message_role,
)


class _AbortSignal:
    def __init__(self) -> None:
        self._event = asyncio.Event()

    @property
    def aborted(self) -> bool:
        return self._event.is_set()


class AbortController:
    def __init__(self) -> None:
        self.signal = _AbortSignal()

    def abort(self) -> None:
        self.signal._event.set()


def default_convert_to_llm(messages: list[AgentMessage]) -> list[Message]:
    return [
        message
        for message in messages
        if message_role(message) in {"user", "assistant", "toolResult"}
    ]


def default_stream_fn(_model: Model, _context: Any, _options: Any):
    raise RuntimeError("No stream function configured. Provide Agent(stream_fn=...) with a provider implementation.")


@dataclass
class AgentOptions:
    initial_state: AgentState | dict[str, Any] | None = None
    convert_to_llm: Callable[[list[AgentMessage]], list[Message] | Any] | None = None
    transform_context: Callable[[list[AgentMessage], Any | None], Any] | None = None
    steering_mode: str = "one-at-a-time"
    follow_up_mode: str = "one-at-a-time"
    stream_fn: Callable[..., Any] | None = None
    session_id: str | None = None
    get_api_key: Callable[[str], Any] | None = None
    thinking_budgets: ThinkingBudgets | None = None
    transport: Transport = "sse"
    max_retry_delay_ms: int | None = None


class Agent:
    def __init__(self, opts: AgentOptions | None = None, **kwargs: Any) -> None:
        options = opts or AgentOptions(**kwargs)

        self._state = AgentState()
        if isinstance(options.initial_state, AgentState):
            self._state = options.initial_state
        elif isinstance(options.initial_state, dict):
            for key, value in options.initial_state.items():
                if hasattr(self._state, key):
                    setattr(self._state, key, value)

        self._listeners: set[Callable[[AgentEvent], None]] = set()
        self._abort_controller: AbortController | None = None
        self._convert_to_llm = options.convert_to_llm or default_convert_to_llm
        self._transform_context = options.transform_context
        self._steering_queue: list[AgentMessage] = []
        self._follow_up_queue: list[AgentMessage] = []
        self._steering_mode = options.steering_mode
        self._follow_up_mode = options.follow_up_mode
        self.stream_fn = options.stream_fn or default_stream_fn
        self._session_id = options.session_id
        self.get_api_key = options.get_api_key
        self._thinking_budgets = options.thinking_budgets
        self._transport: Transport = options.transport
        self._max_retry_delay_ms = options.max_retry_delay_ms
        self._running_prompt: asyncio.Task[None] | None = None

    @property
    def state(self) -> AgentState:
        return self._state

    @property
    def session_id(self) -> str | None:
        return self._session_id

    @session_id.setter
    def session_id(self, value: str | None) -> None:
        self._session_id = value

    @property
    def thinking_budgets(self) -> ThinkingBudgets | None:
        return self._thinking_budgets

    @thinking_budgets.setter
    def thinking_budgets(self, value: ThinkingBudgets | None) -> None:
        self._thinking_budgets = value

    @property
    def transport(self) -> Transport:
        return self._transport

    @property
    def max_retry_delay_ms(self) -> int | None:
        return self._max_retry_delay_ms

    @max_retry_delay_ms.setter
    def max_retry_delay_ms(self, value: int | None) -> None:
        self._max_retry_delay_ms = value

    def subscribe(self, fn: Callable[[AgentEvent], None]) -> Callable[[], None]:
        self._listeners.add(fn)

        def _unsubscribe() -> None:
            self._listeners.discard(fn)

        return _unsubscribe

    def set_system_prompt(self, value: str) -> None:
        self._state.system_prompt = value

    def set_model(self, model: Model) -> None:
        self._state.model = model

    def set_thinking_level(self, level: ThinkingLevel) -> None:
        self._state.thinking_level = level

    def set_steering_mode(self, mode: str) -> None:
        self._steering_mode = mode

    def get_steering_mode(self) -> str:
        return self._steering_mode

    def set_follow_up_mode(self, mode: str) -> None:
        self._follow_up_mode = mode

    def get_follow_up_mode(self) -> str:
        return self._follow_up_mode

    def set_tools(self, tools: list[AgentTool]) -> None:
        self._state.tools = tools

    def set_transport(self, value: Transport) -> None:
        self._transport = value

    def replace_messages(self, messages: list[AgentMessage]) -> None:
        self._state.messages = list(messages)

    def append_message(self, message: AgentMessage) -> None:
        self._state.messages = [*self._state.messages, message]

    def steer(self, message: AgentMessage) -> None:
        self._steering_queue.append(message)

    def follow_up(self, message: AgentMessage) -> None:
        self._follow_up_queue.append(message)

    def clear_steering_queue(self) -> None:
        self._steering_queue = []

    def clear_follow_up_queue(self) -> None:
        self._follow_up_queue = []

    def clear_all_queues(self) -> None:
        self._steering_queue = []
        self._follow_up_queue = []

    def has_queued_messages(self) -> bool:
        return bool(self._steering_queue or self._follow_up_queue)

    def clear_messages(self) -> None:
        self._state.messages = []

    def abort(self) -> None:
        if self._abort_controller:
            self._abort_controller.abort()

    async def wait_for_idle(self) -> None:
        if self._running_prompt:
            await self._running_prompt

    def reset(self) -> None:
        self._state.messages = []
        self._state.is_streaming = False
        self._state.stream_message = None
        self._state.pending_tool_calls = set()
        self._state.error = None
        self.clear_all_queues()

    async def prompt(self, input_message: str | AgentMessage | list[AgentMessage], images: list[ImageContent] | None = None) -> None:
        if self._state.is_streaming:
            raise RuntimeError(
                "Agent is already processing a prompt. Use steer() or follow_up() to queue messages, or wait for completion."
            )

        if isinstance(input_message, list):
            messages = input_message
        elif isinstance(input_message, str):
            content: list[TextContent | ImageContent] = [TextContent(text=input_message)]
            if images:
                content.extend(images)
            messages = [{"role": "user", "content": content, "timestamp": _now_ms()}]
        else:
            messages = [input_message]

        await self._run_loop(messages)

    async def continue_(self) -> None:
        if self._state.is_streaming:
            raise RuntimeError("Agent is already processing. Wait for completion before continuing.")

        messages = self._state.messages
        if not messages:
            raise RuntimeError("No messages to continue from")

        if message_role(messages[-1]) == "assistant":
            queued_steering = self._dequeue_steering_messages()
            if queued_steering:
                await self._run_loop(queued_steering, skip_initial_steering_poll=True)
                return

            queued_follow_up = self._dequeue_follow_up_messages()
            if queued_follow_up:
                await self._run_loop(queued_follow_up)
                return

            raise RuntimeError("Cannot continue from message role: assistant")

        await self._run_loop(None)

    async def _run_loop(self, messages: list[AgentMessage] | None, skip_initial_steering_poll: bool = False) -> None:
        model = self._state.model

        async def _runner() -> None:
            self._abort_controller = AbortController()
            self._state.is_streaming = True
            self._state.stream_message = None
            self._state.error = None

            reasoning = None if self._state.thinking_level == "off" else self._state.thinking_level
            context = AgentContext(
                system_prompt=self._state.system_prompt,
                messages=list(self._state.messages),
                tools=self._state.tools,
            )

            local_skip = skip_initial_steering_poll

            async def _get_steering_messages() -> list[AgentMessage]:
                nonlocal local_skip
                if local_skip:
                    local_skip = False
                    return []
                return self._dequeue_steering_messages()

            config = AgentLoopConfig(
                model=model,
                reasoning=reasoning,
                session_id=self._session_id,
                transport=self._transport,
                thinking_budgets=self._thinking_budgets,
                max_retry_delay_ms=self._max_retry_delay_ms,
                convert_to_llm=self._convert_to_llm,
                transform_context=self._transform_context,
                get_api_key=self.get_api_key,
                get_steering_messages=_get_steering_messages,
                get_follow_up_messages=self._dequeue_follow_up_messages,
            )

            partial: AgentMessage | None = None

            try:
                stream = (
                    agent_loop(messages, context, config, self._abort_controller.signal, self.stream_fn)
                    if messages is not None
                    else agent_loop_continue(context, config, self._abort_controller.signal, self.stream_fn)
                )

                async for event in stream:
                    event_type = getattr(event, "type", "")

                    if event_type in {"message_start", "message_update"}:
                        partial = event.message
                        self._state.stream_message = event.message
                    elif event_type == "message_end":
                        partial = None
                        self._state.stream_message = None
                        self.append_message(event.message)
                    elif event_type == "tool_execution_start":
                        pending = set(self._state.pending_tool_calls)
                        pending.add(event.tool_call_id)
                        self._state.pending_tool_calls = pending
                    elif event_type == "tool_execution_end":
                        pending = set(self._state.pending_tool_calls)
                        pending.discard(event.tool_call_id)
                        self._state.pending_tool_calls = pending
                    elif event_type == "turn_end":
                        if message_role(event.message) == "assistant" and getattr(event.message, "error_message", None):
                            self._state.error = str(getattr(event.message, "error_message"))
                    elif event_type == "agent_end":
                        self._state.is_streaming = False
                        self._state.stream_message = None

                    self._emit(event)

                if partial and message_role(partial) == "assistant":
                    content = getattr(partial, "content", [])
                    only_empty = not any(_has_non_empty_content(part) for part in content)
                    if not only_empty:
                        self.append_message(partial)
                    elif self._abort_controller and self._abort_controller.signal.aborted:
                        raise RuntimeError("Request was aborted")

            except Exception as exc:
                error_message = build_error_message(
                    model,
                    exc,
                    aborted=bool(self._abort_controller and self._abort_controller.signal.aborted),
                )
                self.append_message(error_message)
                self._state.error = str(exc)
                self._emit(AgentEndEvent(messages=[error_message]))
            finally:
                self._state.is_streaming = False
                self._state.stream_message = None
                self._state.pending_tool_calls = set()
                self._abort_controller = None

        task = asyncio.create_task(_runner())
        self._running_prompt = task
        try:
            await task
        finally:
            self._running_prompt = None

    def _dequeue_steering_messages(self) -> list[AgentMessage]:
        if self._steering_mode == "one-at-a-time":
            if self._steering_queue:
                first = self._steering_queue[0]
                self._steering_queue = self._steering_queue[1:]
                return [first]
            return []

        queued = list(self._steering_queue)
        self._steering_queue = []
        return queued

    def _dequeue_follow_up_messages(self) -> list[AgentMessage]:
        if self._follow_up_mode == "one-at-a-time":
            if self._follow_up_queue:
                first = self._follow_up_queue[0]
                self._follow_up_queue = self._follow_up_queue[1:]
                return [first]
            return []

        queued = list(self._follow_up_queue)
        self._follow_up_queue = []
        return queued

    def _emit(self, event: AgentEvent) -> None:
        for listener in list(self._listeners):
            listener(event)

    # TypeScript API aliases for easier migration.
    setSystemPrompt = set_system_prompt
    setModel = set_model
    setThinkingLevel = set_thinking_level
    setSteeringMode = set_steering_mode
    getSteeringMode = get_steering_mode
    setFollowUpMode = set_follow_up_mode
    getFollowUpMode = get_follow_up_mode
    setTools = set_tools
    setTransport = set_transport
    replaceMessages = replace_messages
    appendMessage = append_message
    followUp = follow_up
    clearSteeringQueue = clear_steering_queue
    clearFollowUpQueue = clear_follow_up_queue
    clearAllQueues = clear_all_queues
    hasQueuedMessages = has_queued_messages
    clearMessages = clear_messages
    waitForIdle = wait_for_idle


continue_conversation = Agent.continue_


def _has_non_empty_content(part: Any) -> bool:
    if isinstance(part, dict):
        part_type = part.get("type")
        if part_type == "thinking":
            return bool(str(part.get("thinking", "")).strip())
        if part_type == "text":
            return bool(str(part.get("text", "")).strip())
        if part_type == "toolCall":
            return bool(str(part.get("name", "")).strip())
        return False

    part_type = getattr(part, "type", None)
    if part_type == "thinking":
        return bool(str(getattr(part, "thinking", "")).strip())
    if part_type == "text":
        return bool(str(getattr(part, "text", "")).strip())
    if part_type == "toolCall":
        return bool(str(getattr(part, "name", "")).strip())
    return False


def _now_ms() -> int:
    import time

    return int(time.time() * 1000)




from __future__ import annotations

import asyncio
import inspect
from dataclasses import replace
from typing import Any

from .event_stream import EventStream
from .types import (
    AgentContext,
    AgentEndEvent,
    AgentEvent,
    AgentLoopConfig,
    AgentMessage,
    AgentStartEvent,
    AgentTool,
    AgentToolResult,
    AssistantErrorEvent,
    AssistantMessage,
    AssistantStartEvent,
    Context,
    MessageEndEvent,
    MessageStartEvent,
    MessageUpdateEvent,
    Model,
    TextContent,
    ToolCallContent,
    ToolExecutionEndEvent,
    ToolExecutionStartEvent,
    ToolExecutionUpdateEvent,
    ToolResultMessage,
    TurnEndEvent,
    TurnStartEvent,
    validate_tool_arguments,
)


def agent_loop(
    prompts: list[AgentMessage],
    context: AgentContext,
    config: AgentLoopConfig,
    signal: Any | None = None,
    stream_fn: Any | None = None,
) -> EventStream[AgentEvent, list[AgentMessage]]:
    stream = _create_agent_stream()

    async def _run() -> None:
        new_messages: list[AgentMessage] = list(prompts)
        current_context = AgentContext(
            system_prompt=context.system_prompt,
            messages=[*context.messages, *prompts],
            tools=context.tools,
        )

        try:
            stream.push(AgentStartEvent())
            stream.push(TurnStartEvent())
            for prompt in prompts:
                stream.push(MessageStartEvent(message=prompt))
                stream.push(MessageEndEvent(message=prompt))

            await _run_loop(current_context, new_messages, config, signal, stream, stream_fn)
        except Exception:
            stream.push(AgentEndEvent(messages=new_messages))
            stream.end(new_messages)

    asyncio.create_task(_run())
    return stream


def agent_loop_continue(
    context: AgentContext,
    config: AgentLoopConfig,
    signal: Any | None = None,
    stream_fn: Any | None = None,
) -> EventStream[AgentEvent, list[AgentMessage]]:
    if not context.messages:
        raise ValueError("Cannot continue: no messages in context")

    last = context.messages[-1]
    if _message_role(last) == "assistant":
        raise ValueError("Cannot continue from message role: assistant")

    stream = _create_agent_stream()

    async def _run() -> None:
        new_messages: list[AgentMessage] = []
        current_context = AgentContext(
            system_prompt=context.system_prompt,
            messages=list(context.messages),
            tools=context.tools,
        )

        try:
            stream.push(AgentStartEvent())
            stream.push(TurnStartEvent())
            await _run_loop(current_context, new_messages, config, signal, stream, stream_fn)
        except Exception:
            stream.push(AgentEndEvent(messages=new_messages))
            stream.end(new_messages)

    asyncio.create_task(_run())
    return stream


def _create_agent_stream() -> EventStream[AgentEvent, list[AgentMessage]]:
    return EventStream[AgentEvent, list[AgentMessage]](
        lambda event: getattr(event, "type", "") == "agent_end",
        lambda event: event.messages if isinstance(event, AgentEndEvent) else [],
    )


async def _run_loop(
    current_context: AgentContext,
    new_messages: list[AgentMessage],
    config: AgentLoopConfig,
    signal: Any | None,
    stream: EventStream[AgentEvent, list[AgentMessage]],
    stream_fn: Any | None,
) -> None:
    first_turn = True
    pending_messages = await _get_messages(config.get_steering_messages)

    while True:
        has_more_tool_calls = True
        steering_after_tools: list[AgentMessage] | None = None

        while has_more_tool_calls or pending_messages:
            if not first_turn:
                stream.push(TurnStartEvent())
            else:
                first_turn = False

            if pending_messages:
                for message in pending_messages:
                    stream.push(MessageStartEvent(message=message))
                    stream.push(MessageEndEvent(message=message))
                    current_context.messages.append(message)
                    new_messages.append(message)
                pending_messages = []

            message = await _stream_assistant_response(current_context, config, signal, stream, stream_fn)
            new_messages.append(message)

            if message.stop_reason in ("error", "aborted"):
                stream.push(TurnEndEvent(message=message, tool_results=[]))
                stream.push(AgentEndEvent(messages=new_messages))
                stream.end(new_messages)
                return

            tool_calls = [part for part in message.content if _is_tool_call(part)]
            has_more_tool_calls = len(tool_calls) > 0

            tool_results: list[ToolResultMessage] = []
            if has_more_tool_calls:
                tool_execution = await _execute_tool_calls(
                    current_context.tools,
                    message,
                    signal,
                    stream,
                    config.get_steering_messages,
                )
                tool_results.extend(tool_execution["tool_results"])
                steering_after_tools = tool_execution.get("steering_messages")

                for result in tool_results:
                    current_context.messages.append(result)
                    new_messages.append(result)

            stream.push(TurnEndEvent(message=message, tool_results=tool_results))

            if steering_after_tools:
                pending_messages = steering_after_tools
                steering_after_tools = None
            else:
                pending_messages = await _get_messages(config.get_steering_messages)

        follow_up_messages = await _get_messages(config.get_follow_up_messages)
        if follow_up_messages:
            pending_messages = follow_up_messages
            continue

        break

    stream.push(AgentEndEvent(messages=new_messages))
    stream.end(new_messages)


async def _stream_assistant_response(
    context: AgentContext,
    config: AgentLoopConfig,
    signal: Any | None,
    stream: EventStream[AgentEvent, list[AgentMessage]],
    stream_fn: Any | None,
) -> AssistantMessage:
    if not config.model:
        raise ValueError("AgentLoopConfig.model is required")
    if not config.convert_to_llm:
        raise ValueError("AgentLoopConfig.convert_to_llm is required")

    messages = context.messages
    if config.transform_context:
        transformed = config.transform_context(messages, signal)
        messages = await _maybe_await(transformed)

    llm_messages = await _maybe_await(config.convert_to_llm(messages))

    llm_context = Context(
        system_prompt=context.system_prompt,
        messages=llm_messages,
        tools=context.tools,
    )

    selected_stream_fn = stream_fn
    if selected_stream_fn is None:
        raise RuntimeError("No stream function provided. Set Agent.stream_fn or pass stream_fn to agent_loop().")

    resolved_api_key = config.api_key
    if config.get_api_key:
        resolved_api_key = await _maybe_await(config.get_api_key(config.model.provider)) or config.api_key

    options = replace(config)
    options.api_key = resolved_api_key
    options.signal = signal

    response_candidate = selected_stream_fn(config.model, llm_context, options)
    response = await _maybe_await(response_candidate)

    partial_message: AssistantMessage | None = None
    added_partial = False

    async for event in response:
        event_type = getattr(event, "type", "")

        if event_type == "start":
            if not isinstance(event, AssistantStartEvent):
                partial_message = getattr(event, "partial")
            else:
                partial_message = event.partial
            context.messages.append(partial_message)
            added_partial = True
            stream.push(MessageStartEvent(message=partial_message))
            continue

        if event_type in {
            "text_start",
            "text_delta",
            "text_end",
            "thinking_start",
            "thinking_delta",
            "thinking_end",
            "toolcall_start",
            "toolcall_delta",
            "toolcall_end",
        }:
            if partial_message is not None:
                partial_message = getattr(event, "partial", partial_message)
                context.messages[-1] = partial_message
                stream.push(
                    MessageUpdateEvent(
                        assistant_message_event=event,
                        message=partial_message,
                    )
                )
            continue

        if event_type in {"done", "error"}:
            final_message = await response.result()
            if added_partial:
                context.messages[-1] = final_message
            else:
                context.messages.append(final_message)
                stream.push(MessageStartEvent(message=final_message))
            stream.push(MessageEndEvent(message=final_message))
            return final_message

    return await response.result()


async def _execute_tool_calls(
    tools: list[AgentTool] | None,
    assistant_message: AssistantMessage,
    signal: Any | None,
    stream: EventStream[AgentEvent, list[AgentMessage]],
    get_steering_messages: Any | None,
) -> dict[str, Any]:
    tool_calls = [_to_tool_call(part) for part in assistant_message.content if _is_tool_call(part)]
    results: list[ToolResultMessage] = []
    steering_messages: list[AgentMessage] | None = None

    for index, tool_call in enumerate(tool_calls):
        tool = None
        if tools:
            tool = next((candidate for candidate in tools if candidate.name == tool_call.name), None)

        stream.push(
            ToolExecutionStartEvent(
                tool_call_id=tool_call.id,
                tool_name=tool_call.name,
                args=tool_call.arguments,
            )
        )

        is_error = False
        try:
            if tool is None:
                raise RuntimeError(f"Tool {tool_call.name} not found")

            validated_args = validate_tool_arguments(tool, tool_call)

            def _on_update(partial_result: AgentToolResult) -> None:
                stream.push(
                    ToolExecutionUpdateEvent(
                        tool_call_id=tool_call.id,
                        tool_name=tool_call.name,
                        args=tool_call.arguments,
                        partial_result=partial_result,
                    )
                )

            result = await tool.execute(tool_call.id, validated_args, signal, _on_update)
        except Exception as exc:
            result = AgentToolResult(
                content=[TextContent(text=str(exc))],
                details={},
            )
            is_error = True

        stream.push(
            ToolExecutionEndEvent(
                tool_call_id=tool_call.id,
                tool_name=tool_call.name,
                result=result,
                is_error=is_error,
            )
        )

        tool_result_message = ToolResultMessage(
            tool_call_id=tool_call.id,
            tool_name=tool_call.name,
            content=result.content,
            details=result.details,
            is_error=is_error,
        )

        results.append(tool_result_message)
        stream.push(MessageStartEvent(message=tool_result_message))
        stream.push(MessageEndEvent(message=tool_result_message))

        if get_steering_messages:
            steering = await _get_messages(get_steering_messages)
            if steering:
                steering_messages = steering
                for skipped in tool_calls[index + 1 :]:
                    results.append(_skip_tool_call(skipped, stream))
                break

    return {
        "tool_results": results,
        "steering_messages": steering_messages,
    }


def _skip_tool_call(
    tool_call: ToolCallContent,
    stream: EventStream[AgentEvent, list[AgentMessage]],
) -> ToolResultMessage:
    result = AgentToolResult(
        content=[TextContent(text="Skipped due to queued user message.")],
        details={},
    )

    stream.push(
        ToolExecutionStartEvent(
            tool_call_id=tool_call.id,
            tool_name=tool_call.name,
            args=tool_call.arguments,
        )
    )
    stream.push(
        ToolExecutionEndEvent(
            tool_call_id=tool_call.id,
            tool_name=tool_call.name,
            result=result,
            is_error=True,
        )
    )

    message = ToolResultMessage(
        tool_call_id=tool_call.id,
        tool_name=tool_call.name,
        content=result.content,
        details={},
        is_error=True,
    )
    stream.push(MessageStartEvent(message=message))
    stream.push(MessageEndEvent(message=message))
    return message


async def _get_messages(callback: Any | None) -> list[AgentMessage]:
    if not callback:
        return []
    result = callback()
    messages = await _maybe_await(result)
    return list(messages or [])


async def _maybe_await(value: Any) -> Any:
    if inspect.isawaitable(value):
        return await value
    return value


def _is_tool_call(part: Any) -> bool:
    if isinstance(part, ToolCallContent):
        return True
    if isinstance(part, dict):
        return part.get("type") == "toolCall"
    return getattr(part, "type", None) == "toolCall"


def _to_tool_call(part: Any) -> ToolCallContent:
    if isinstance(part, ToolCallContent):
        return part
    if isinstance(part, dict):
        return ToolCallContent(
            id=str(part.get("id", "")),
            name=str(part.get("name", "")),
            arguments=dict(part.get("arguments", {})),
        )
    return ToolCallContent(
        id=str(getattr(part, "id", "")),
        name=str(getattr(part, "name", "")),
        arguments=dict(getattr(part, "arguments", {})),
    )


def _message_role(message: AgentMessage) -> str:
    if isinstance(message, dict):
        return str(message.get("role", ""))
    return str(getattr(message, "role", ""))


def build_error_message(model: Model, error: Exception | str, aborted: bool = False) -> AssistantMessage:
    return AssistantMessage(
        content=[TextContent(text="")],
        api=model.api,
        provider=model.provider,
        model=model.id,
        stop_reason="aborted" if aborted else "error",
        error_message=str(error),
    )


def as_error_event(message: AssistantMessage) -> AssistantErrorEvent:
    return AssistantErrorEvent(reason="aborted" if message.stop_reason == "aborted" else "error", error=message)




from __future__ import annotations

import asyncio
import json
from dataclasses import asdict, dataclass, is_dataclass
from typing import Any

import httpx

from .event_stream import EventStream
from .types import (
    AssistantDoneEvent,
    AssistantErrorEvent,
    AssistantMessage,
    AssistantMessageEvent,
    AssistantStartEvent,
    AssistantTextDeltaEvent,
    AssistantTextEndEvent,
    AssistantTextStartEvent,
    AssistantThinkingDeltaEvent,
    AssistantThinkingEndEvent,
    AssistantThinkingStartEvent,
    AssistantToolCallDeltaEvent,
    AssistantToolCallEndEvent,
    AssistantToolCallStartEvent,
    Context,
    Model,
    SimpleStreamOptions,
    StopReason,
    TextContent,
    ThinkingContent,
    ToolCallContent,
    Usage,
    UsageCost,
)


class ProxyError(RuntimeError):
    pass


class ProxyMessageEventStream(EventStream[AssistantMessageEvent, AssistantMessage]):
    def __init__(self) -> None:
        super().__init__(
            lambda event: getattr(event, "type", "") in {"done", "error"},
            lambda event: event.message if getattr(event, "type", "") == "done" else event.error,
        )


@dataclass
class ProxyStreamOptions(SimpleStreamOptions):
    auth_token: str = ""
    proxy_url: str = ""


def stream_proxy(model: Model, context: Context, options: ProxyStreamOptions) -> ProxyMessageEventStream:
    stream = ProxyMessageEventStream()

    async def _run() -> None:
        partial = AssistantMessage(
            content=[],
            api=model.api,
            provider=model.provider,
            model=model.id,
            usage=Usage(cost=UsageCost()),
        )
        partial_json: dict[int, str] = {}

        try:
            if not options.proxy_url:
                raise ProxyError("proxy_url is required")
            if not options.auth_token:
                raise ProxyError("auth_token is required")

            payload = {
                "model": _serialize(model),
                "context": _serialize_context(context),
                "options": {
                    "temperature": options.temperature,
                    "maxTokens": options.max_tokens,
                    "reasoning": options.reasoning,
                },
            }

            async with httpx.AsyncClient(timeout=None) as client:
                async with client.stream(
                    "POST",
                    f"{options.proxy_url.rstrip('/')}/api/stream",
                    headers={
                        "Authorization": f"Bearer {options.auth_token}",
                        "Content-Type": "application/json",
                    },
                    json=payload,
                ) as response:
                    if response.status_code >= 400:
                        message = f"Proxy error: {response.status_code}"
                        try:
                            data = await response.aread()
                            parsed = json.loads(data.decode("utf-8"))
                            if isinstance(parsed, dict) and parsed.get("error"):
                                message = f"Proxy error: {parsed['error']}"
                        except Exception:
                            pass
                        raise ProxyError(message)

                    async for line in response.aiter_lines():
                        if options.signal and options.signal.aborted:
                            raise ProxyError("Request aborted by user")
                        if not line.startswith("data: "):
                            continue

                        data = line[6:].strip()
                        if not data:
                            continue

                        proxy_event = json.loads(data)
                        event = _process_proxy_event(proxy_event, partial, partial_json)
                        if event:
                            stream.push(event)

            if options.signal and options.signal.aborted:
                raise ProxyError("Request aborted by user")

            stream.end()
        except Exception as exc:
            reason: StopReason = "aborted" if options.signal and options.signal.aborted else "error"
            partial.stop_reason = reason
            partial.error_message = str(exc)
            stream.push(AssistantErrorEvent(reason="aborted" if reason == "aborted" else "error", error=partial))
            stream.end()

    asyncio.create_task(_run())
    return stream


def _process_proxy_event(
    proxy_event: dict[str, Any],
    partial: AssistantMessage,
    partial_json: dict[int, str],
) -> AssistantMessageEvent | None:
    event_type = proxy_event.get("type")

    if event_type == "start":
        return AssistantStartEvent(partial=partial)

    if event_type == "text_start":
        index = int(proxy_event["contentIndex"])
        _set_content(partial, index, TextContent(text=""))
        return AssistantTextStartEvent(content_index=index, partial=partial)

    if event_type == "text_delta":
        index = int(proxy_event["contentIndex"])
        content = _get_content(partial, index)
        if not isinstance(content, TextContent):
            raise ProxyError("Received text_delta for non-text content")
        content.text += str(proxy_event.get("delta", ""))
        return AssistantTextDeltaEvent(content_index=index, delta=str(proxy_event.get("delta", "")), partial=partial)

    if event_type == "text_end":
        index = int(proxy_event["contentIndex"])
        content = _get_content(partial, index)
        if not isinstance(content, TextContent):
            raise ProxyError("Received text_end for non-text content")
        return AssistantTextEndEvent(content_index=index, content=content.text, partial=partial)

    if event_type == "thinking_start":
        index = int(proxy_event["contentIndex"])
        _set_content(partial, index, ThinkingContent(thinking=""))
        return AssistantThinkingStartEvent(content_index=index, partial=partial)

    if event_type == "thinking_delta":
        index = int(proxy_event["contentIndex"])
        content = _get_content(partial, index)
        if not isinstance(content, ThinkingContent):
            raise ProxyError("Received thinking_delta for non-thinking content")
        content.thinking += str(proxy_event.get("delta", ""))
        return AssistantThinkingDeltaEvent(content_index=index, delta=str(proxy_event.get("delta", "")), partial=partial)

    if event_type == "thinking_end":
        index = int(proxy_event["contentIndex"])
        content = _get_content(partial, index)
        if not isinstance(content, ThinkingContent):
            raise ProxyError("Received thinking_end for non-thinking content")
        signature = proxy_event.get("contentSignature")
        if signature:
            content.thinking_signature = str(signature)
        return AssistantThinkingEndEvent(content_index=index, content=content.thinking, partial=partial)

    if event_type == "toolcall_start":
        index = int(proxy_event["contentIndex"])
        tool_call = ToolCallContent(
            id=str(proxy_event.get("id", "")),
            name=str(proxy_event.get("toolName", "")),
            arguments={},
        )
        _set_content(partial, index, tool_call)
        partial_json[index] = ""
        return AssistantToolCallStartEvent(content_index=index, partial=partial)

    if event_type == "toolcall_delta":
        index = int(proxy_event["contentIndex"])
        content = _get_content(partial, index)
        if not isinstance(content, ToolCallContent):
            raise ProxyError("Received toolcall_delta for non-toolCall content")
        delta = str(proxy_event.get("delta", ""))
        partial_json[index] = partial_json.get(index, "") + delta
        parsed = _parse_streaming_json(partial_json[index])
        if isinstance(parsed, dict):
            content.arguments = parsed
        return AssistantToolCallDeltaEvent(content_index=index, delta=delta, partial=partial)

    if event_type == "toolcall_end":
        index = int(proxy_event["contentIndex"])
        content = _get_content(partial, index)
        if not isinstance(content, ToolCallContent):
            return None
        partial_json.pop(index, None)
        return AssistantToolCallEndEvent(content_index=index, tool_call=content, partial=partial)

    if event_type == "done":
        partial.stop_reason = str(proxy_event.get("reason", "stop"))  # type: ignore[assignment]
        partial.usage = _deserialize_usage(proxy_event.get("usage"))
        return AssistantDoneEvent(reason=partial.stop_reason if partial.stop_reason in {"stop", "length", "toolUse"} else "stop", message=partial)

    if event_type == "error":
        reason = str(proxy_event.get("reason", "error"))
        partial.stop_reason = "aborted" if reason == "aborted" else "error"
        partial.error_message = proxy_event.get("errorMessage")
        partial.usage = _deserialize_usage(proxy_event.get("usage"))
        return AssistantErrorEvent(reason="aborted" if reason == "aborted" else "error", error=partial)

    return None


def _serialize_context(context: Context) -> dict[str, Any]:
    return {
        "systemPrompt": context.system_prompt,
        "messages": [_serialize(message) for message in context.messages],
        # Tool execution happens client-side in this package.
        "tools": None,
    }


def _serialize(value: Any) -> Any:
    if is_dataclass(value):
        return asdict(value)
    if isinstance(value, list):
        return [_serialize(item) for item in value]
    if isinstance(value, dict):
        return {key: _serialize(item) for key, item in value.items()}
    return value


def _deserialize_usage(value: Any) -> Usage:
    if not isinstance(value, dict):
        return Usage(cost=UsageCost())

    cost = value.get("cost", {}) if isinstance(value.get("cost"), dict) else {}
    return Usage(
        input=int(value.get("input", 0)),
        output=int(value.get("output", 0)),
        cache_read=int(value.get("cacheRead", value.get("cache_read", 0))),
        cache_write=int(value.get("cacheWrite", value.get("cache_write", 0))),
        total_tokens=int(value.get("totalTokens", value.get("total_tokens", 0))),
        cost=UsageCost(
            input=float(cost.get("input", 0)),
            output=float(cost.get("output", 0)),
            cache_read=float(cost.get("cacheRead", cost.get("cache_read", 0))),
            cache_write=float(cost.get("cacheWrite", cost.get("cache_write", 0))),
            total=float(cost.get("total", 0)),
        ),
    )


def _set_content(partial: AssistantMessage, index: int, content: Any) -> None:
    while len(partial.content) <= index:
        partial.content.append(TextContent(text=""))
    partial.content[index] = content


def _get_content(partial: AssistantMessage, index: int) -> Any:
    if index >= len(partial.content):
        raise ProxyError(f"Content index out of range: {index}")
    return partial.content[index]


def _parse_streaming_json(value: str) -> dict[str, Any] | None:
    try:
        parsed = json.loads(value)
    except json.JSONDecodeError:
        return None
    if isinstance(parsed, dict):
        return parsed
    return None



from __future__ import annotations

import asyncio
from collections.abc import AsyncIterator, Callable
from typing import Generic, TypeVar

EventT = TypeVar("EventT")
ResultT = TypeVar("ResultT")


class EventStream(Generic[EventT, ResultT], AsyncIterator[EventT]):
    """A push-based async event stream with terminal result extraction."""

    def __init__(
        self,
        is_terminal_event: Callable[[EventT], bool],
        terminal_result: Callable[[EventT], ResultT],
    ) -> None:
        self._is_terminal_event = is_terminal_event
        self._terminal_result = terminal_result
        self._queue: asyncio.Queue[EventT | None] = asyncio.Queue()
        self._result_future: asyncio.Future[ResultT] = asyncio.get_running_loop().create_future()
        self._closed = False

    def push(self, event: EventT) -> None:
        if self._closed:
            return
        self._queue.put_nowait(event)

        if self._is_terminal_event(event) and not self._result_future.done():
            self._result_future.set_result(self._terminal_result(event))
            self._closed = True
            self._queue.put_nowait(None)

    def end(self, result: ResultT | None = None) -> None:
        if self._closed:
            return
        self._closed = True
        if result is not None and not self._result_future.done():
            self._result_future.set_result(result)
        elif result is None and not self._result_future.done():
            self._result_future.set_exception(RuntimeError("Event stream ended before terminal event"))
        self._queue.put_nowait(None)

    async def result(self) -> ResultT:
        return await self._result_future

    def __aiter__(self) -> EventStream[EventT, ResultT]:
        return self

    async def __anext__(self) -> EventT:
        item = await self._queue.get()
        if item is None:
            raise StopAsyncIteration
        return item

